/**
 * 3d secure authentication using Pci Bookings Orchestra service
 */
var ts3DSecure = {
    /**
     * The web socket session with the ePaytools service
     */
    webSocket : null,

    /**
     * The session token as generated by ePayTools
     */
    sessionToken: null,

    /**
     * iFrame for challenge/fingerprint
     */
    iFrame : null,

    /**
     * The callback function to be invoked upon completion of the 3DS authentication
     */
    callback : null,

    /**
     * Base url for Wbsocket
     */
    wsBaseUrl : "wss://3ds.epaytools.com",

    /**
     * Base url for Api
     */
    apiBaseUrl : "https://api.epaytools.com",

    /**
     * Main function to perform a 3d secure authentication session
     * 
     * @param {string} sessionToken - token obtained from the Orchestra service
     * @param {string} iFrameElementId - Id of element in which a challenge iFrame is to be created
     * @param {Object} cardData - The card details
     * @param callback - the callback function to be invoked upon completion
     */

    perform3ds: function(sessionToken, iFrameElementId, cardData, callback) {
        if (!iFrameElementId || !cardData  || !sessionToken || !callback) {
            console.error("mainContentId || iFrameElementId or cardData or sessionToken or callback are empty; quitting");
            return;
        }

        this.iFrame = document.getElementById(iFrameElementId);
        if (!this.iFrame) {
            console.error(iFrameElementId + " not accessible");
            return;
        }

        // store callback function + sessionToken
        this.callback = callback;
        this.sessionToken = sessionToken;

        // Connect to WebSocket
        this.openWebSocket(function () {
            // Upon open completion, send a message with the session token
            ts3DSecure.sendWebSocketMessage(ts3DSecure.sessionToken);

            // set timer on receiving a response from server
            var receiveTimer = setTimeout(()=>{ts3DSecure.handleError("Timeout on receiving websocket ECHO response","websocket ECHO")}, 20000);

            // Bind the on message received event to accept the Jws with additional data
            ts3DSecure.webSocket.onmessage = function (message) {
                // clear receive message timer
                clearTimeout(receiveTimer);
                // store received session token (received as a json string)
                ts3DSecure.sessionToken = JSON.parse(message.data);
                // Init 3DS session 
                ts3DSecure.init3DSecAuthSession(cardData);
            };
        });
    },

    /**
     * init 3DS Authentication Session
     * @param {*} modelData - containing card info
     */
    init3DSecAuthSession: function (modelData) {
        // Extend card model with browser/screen/window properties 
        modelData.windowWidth = self.innerWidth;
        modelData.windowHeight = self.innerHeight;
        modelData.timeZoneOffsetMinutes = new Date().getTimezoneOffset();
        modelData.colorDepth = screen.colorDepth;
        modelData.browserScreenWidth = screen.availWidth;
        modelData.browserScreenHeight = screen.availHeight;
        modelData.browserJavaEnabled = navigator.BrowserJavaEnabled;
        var data = JSON.stringify(modelData);
        console.log(data);
      
        const xhr = new XMLHttpRequest();

        xhr.onload = function () {
            // Request finished. Do processing here.
            var response = JSON.parse(this.responseText);
            if (xhr.status < 300)
            // Success
                ts3DSecure.handleInitSessionResponse(response);
            else
            // Failure
                ts3DSecure.handleError(this.responseText,"init3DSecAuthSession Error");
        };
          
        xhr.ontimeout = function (e) {
            // XMLHttpRequest timed out
            ts3DSecure.handleError("Http post request timed out","init3DSecAuthSession")        ;
        };

        xhr.open("POST", this.apiBaseUrl + "/threeDsSessions/init", true);
        xhr.timeout = 20000; // 20 sec. (time in milliseconds)
        xhr.setRequestHeader("Content-type", "application/json");
        xhr.setRequestHeader("Authorization", "Bearer " + this.sessionToken);
        xhr.send(data);
    },

    /**
     * Open a Web socket connection
     * @callback onOpenCallback - the callback function to be invoked upon websocket connection
     */
    openWebSocket: function (onOpenCallback) {
        // construct Websocket Url
        var webSocketURL = this.wsBaseUrl + "?token=" + this.sessionToken;

        // Initiate a websocket session and pass the sessionToken as parameter in querystring. 
        // ways to add headers are not available.Pssing token in the path is not supported by AWS ApiGateway
        this.webSocket = new WebSocket(webSocketURL);
        this.webSocket.addEventListener('error', function (event) {
            ts3DSecure.handleError(event, 'WebSocket Open');
        });
        this.webSocket.onopen = onOpenCallback;
    },

    /**
     * Send a message over web socket
     * 
     * @param {string} message 
     */
    sendWebSocketMessage: function (message) {
        if (!this.webSocket) {
            console.error("Web socket is not open");
            return;
        }
        if (!message) {
            console.error("Message is empty; nothing to send");
            return;
        }
        this.webSocket.send(message);
    },

    /**
     * Handle response for init session
     * 
     * @param {Object} message 
     */
    handleInitSessionResponse: function(message)
    {
        console.log(' init: Recieved ' + message.clientMessage.messageType + ' message from server');
        switch (message.clientMessage.messageType.toLowerCase()) {
            case 'technicalproblem':
            case 'rejected':
            case 'authenticated':
            case 'cardnotenrolled':
                // no more action required;
                this.callback(message);
                break;

            case 'challengerequired':
                // Store new token
                ts3DSecure.sessionToken = message.token;
                this.performChallenge(message.clientMessage.ascUri);         
                break;

            case 'fingerprintrequired':
                // Store new token
                ts3DSecure.sessionToken = message.token;

                // Set timer on overall Fingerprint flow
                var fingerPrintTimeout = setTimeout(()=>{
                    ts3DSecure.handleError("Timeout on receiving websocket response after creating a fingerprint iFrame","fingerprintrequired")
                    // try obtaining fingerprint status even if notification has not been received
                    ts3DSecure.getFingerPrintResults(false);

                }, 20000);

                console.log("Creating a fingerprint iFrame");

                // set Websocket on receive challenge results
                this.webSocket.onmessage = function(event) {
                    // Clear fingerPrint Timeout
                    clearTimeout(fingerPrintTimeout);

                    // Handle the response message - get status
                    ts3DSecure.getFingerPrintResults(true);
                };

                var ifrm = document.createElement("iframe");
                ifrm.style.display = "none";
                document.body.appendChild(ifrm);

                // inject form into iFrame, then perform a submit operation
                var body =
                "<body onload='document.getElementById(\"frmRedirect\").submit();'>" +
                "<form method='post' id='frmRedirect' action='" + message.clientMessage.fingerPrintUri + "'><input type='hidden' name='threeDSMethodData' value='" + message.clientMessage.fingerPrintData + "' /><input type='submit'/></form>" +
                "</body>";
 
                var doc = ifrm.contentWindow.document;
                doc.open();
                doc.write(body);
                doc.close();
       
                break;
    
            default:
                console.error('Received unknown message - ' + message.MessageType);
        }
    },

    /**
     * Perform challenge. Can be called either after session initiation or after a fingerprint test
     * @param {string} ascUri - the URI to use for iFrame src tag.
     */
    performChallenge: function (ascUri) {   
        var challengeTimeout = setTimeout(()=>{ts3DSecure.handleError("Timeout on receiving websocket response after creating a challenge iFrame","challengerequired")}, 600000 /* 10 min. */);

        // set Websocket on receive challenge results
        this.webSocket.onmessage = function(event) {
            // Clear challenge Timeout
            clearTimeout(challengeTimeout);

            // destroy the iframe
            ts3DSecure.iFrame.style.display = "none";

            // Handle the response message
            ts3DSecure.handleServerResultMessage(JSON.parse(event.data));
        };

        // Open iframe with source to challenge over the entire existing window
        console.log("Creating a challenge iFrame");

        this.iFrame.style.display = "inline";
        this.iFrame.setAttribute("src", ascUri);
    },


    /**
     * Get fingerprint results
     * 
     * @param {bool} isNotified - is this called after being notified on completion (otherwise - after a timeout)
     */
    getFingerPrintResults: function(isNotified) {
        // Send a request to get results
        const xhr = new XMLHttpRequest();
        xhr.onload = function() {
            if (xhr.status < 300) {
                var message = JSON.parse(this.responseText);
                if (message.messageType.toLowerCase() == 'challengerequired')
                    // challenge required
                    ts3DSecure.performChallenge(message.ascUri);
                else
                    // final status
                    ts3DSecure.handleServerResultMessage(message)
            }
            else {
                ts3DSecure.handleError(this.responseText,"handleFingerPrintCompleteMessage error")
            }
        };

        xhr.ontimeout = function (e) {
            // XMLHttpRequest timed out.
            ts3DSecure.handleError("Http get request timed out","handleFingerPrintCompleteMessage")        ;
        };

        xhr.open("GET", ts3DSecure.apiBaseUrl + "/threeDsSessions/fingerPrintStatus?isNotified=" + isNotified, true);
        xhr.timeout = 20000; // 20 sec
        xhr.setRequestHeader("Authorization", "Bearer " + this.sessionToken);
        xhr.send();
    },

    /**
     * Handle Server result Message
     * 
     * @param {*} message 
     */
    handleServerResultMessage: function (message) {
        console.log('3DS: Received ' + message.messageType + ' message from server');
        this.callback(message);
    },

    /**
     * Error reporting
     * @param {string} reason
     * @param {string} ref
     */
    handleError: function name(reason, ref) {
        console.error(ref + ': ' + reason);
        var errMessage = {
            messageType :'TechnicalProblem',
            reason: reason,
            reference: ref   
        };
        this.callback(errMessage);
    }
};